@@ -1,5 +1,5 @@
 /*
- * (C) 2019 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
+ * (C) 2019-2025 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
  * All Rights Reserved
  *
  * SPDX-License-Identifier: AGPL-3.0+
@@ -23,6 +23,7 @@
 #include <osmocom/core/linuxlist.h>
 #include <osmocom/core/logging.h>
 #include <osmocom/core/fsm.h>
+#include <osmocom/core/stat_item.h>
 #include <osmocom/gsm/bssmap_le.h>
 #include <osmocom/sigtran/sccp_helpers.h>
 
@@ -57,6 +58,7 @@ static struct lb_peer *lb_peer_alloc(struct sccp_lb_inst *sli, const struct osmo
 	fi->priv = lbp;
 
 	llist_add(&lbp->entry, &sli->lb_peers);
+	osmo_stat_item_inc(osmo_stat_item_group_get_item(g_smlc->statg, SMLC_STAT_LB_PEERS_TOTAL), 1);
 
 	return lbp;
 }
@@ -69,7 +71,7 @@ struct lb_peer *lb_peer_find_or_create(struct sccp_lb_inst *sli, const struct os
 	return lb_peer_alloc(sli, peer_addr);
 }
 
-struct lb_peer *lb_peer_find(struct sccp_lb_inst *sli, const struct osmo_sccp_addr *peer_addr)
+struct lb_peer *lb_peer_find(const struct sccp_lb_inst *sli, const struct osmo_sccp_addr *peer_addr)
 {
 	struct lb_peer *lbp;
 	llist_for_each_entry(lbp, &sli->lb_peers, entry) {
@@ -80,6 +82,15 @@ struct lb_peer *lb_peer_find(struct sccp_lb_inst *sli, const struct osmo_sccp_ad
 	return NULL;
 }
 
+/* Find an lb_peer by its remote SCCP address */
+struct lb_peer *lb_peer_find_by_pc(const struct sccp_lb_inst *sli, uint32_t pc)
+{
+	struct osmo_sccp_addr rem_addr;
+
+	osmo_sccp_make_addr_pc_ssn(&rem_addr, pc, OSMO_SCCP_SSN_BSC_BSSAP_LE);
+	return lb_peer_find(sli, &rem_addr);
+}
+
 static const struct osmo_tdef_state_timeout lb_peer_fsm_timeouts[32] = {
 	[LB_PEER_ST_WAIT_RX_RESET_ACK] = { .T = -13 },
 	[LB_PEER_ST_DISCARDING] = { .T = -14 },
@@ -88,7 +99,7 @@ static const struct osmo_tdef_state_timeout lb_peer_fsm_timeouts[32] = {
 #define lb_peer_state_chg(LB_PEER, NEXT_STATE) \
 	osmo_tdef_fsm_inst_state_chg((LB_PEER)->fi, NEXT_STATE, lb_peer_fsm_timeouts, g_smlc_tdefs, 5)
 
-void lb_peer_discard_all_conns(struct lb_peer *lbp)
+static void lb_peer_discard_all_conns(struct lb_peer *lbp)
 {
 	struct lb_conn *lb_conn, *next;
 
@@ -137,7 +148,7 @@ static void lb_peer_rx_reset_ack(struct lb_peer *lbp, struct msgb* msg)
 	lb_peer_state_chg(lbp, LB_PEER_ST_READY);
 }
 
-void lb_peer_reset(struct lb_peer *lbp)
+static void lb_peer_reset(struct lb_peer *lbp)
 {
 	struct bssap_le_pdu reset = {
 		.discr = BSSAP_LE_MSG_DISCR_BSSMAP_LE,
@@ -167,7 +178,18 @@ void lb_peer_reset(struct lb_peer *lbp)
 	}
 }
 
-void lb_peer_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+static void lb_peer_disconnect(struct sccp_lb_inst *sli, uint32_t conn_id)
+{
+	struct lb_conn *lb_conn;
+	llist_for_each_entry(lb_conn, &sli->lb_conns, entry) {
+		if (lb_conn->sccp_conn_id == conn_id) {
+			lb_conn_discard(lb_conn);
+			return;
+		}
+	}
+}
+
+static void lb_peer_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct lb_peer *lbp = fi->priv;
 	struct lb_peer_ev_ctx *ctx = data;
@@ -196,7 +218,7 @@ void lb_peer_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	}
 }
 
-void lb_peer_st_wait_rx_reset(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+static void lb_peer_st_wait_rx_reset(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct lb_peer *lbp = fi->priv;
 	struct lb_peer_ev_ctx *ctx;
@@ -224,13 +246,22 @@ void lb_peer_st_wait_rx_reset(struct osmo_fsm_inst *fi, uint32_t event, void *da
 		lb_peer_rx_reset(lbp, msg);
 		return;
 
+	case LB_PEER_EV_AVAILABLE:
+		/* Send a RESET to the peer. */
+		lb_peer_reset(lbp);
+		return;
+
+	case LB_PEER_EV_UNAVAILABLE:
+		/* Do nothing, wait for peer to come up again. */
+		return;
+
 	default:
 		LOG_LB_PEER(lbp, LOGL_ERROR, "Unhandled event: %s\n", osmo_fsm_event_name(&lb_peer_fsm, event));
 		return;
 	}
 }
 
-void lb_peer_st_wait_rx_reset_ack(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+static void lb_peer_st_wait_rx_reset_ack(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct lb_peer *lbp = fi->priv;
 	struct lb_peer_ev_ctx *ctx;
@@ -259,13 +290,28 @@ void lb_peer_st_wait_rx_reset_ack(struct osmo_fsm_inst *fi, uint32_t event, void
 		lb_peer_rx_reset(lbp, msg);
 		return;
 
+	case LB_PEER_EV_AVAILABLE:
+		/* Send a RESET to the peer. */
+		lb_peer_reset(lbp);
+		return;
+
+	case LB_PEER_EV_UNAVAILABLE:
+		lb_peer_state_chg(lbp, LB_PEER_ST_WAIT_RX_RESET);
+		return;
+
 	default:
 		LOG_LB_PEER(lbp, LOGL_ERROR, "Unhandled event: %s\n", osmo_fsm_event_name(&lb_peer_fsm, event));
 		return;
 	}
 }
 
-void lb_peer_st_ready(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+static void lb_peer_st_ready_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
+	if (prev_state != LB_PEER_ST_READY)
+		osmo_stat_item_inc(osmo_stat_item_group_get_item(g_smlc->statg, SMLC_STAT_LB_PEERS_ACTIVE), 1);
+}
+
+static void lb_peer_st_ready(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct lb_peer *lbp = fi->priv;
 	struct lb_peer_ev_ctx *ctx;
@@ -323,12 +369,28 @@ void lb_peer_st_ready(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 		lb_peer_rx_reset(lbp, msg);
 		return;
 
+	case LB_PEER_EV_AVAILABLE:
+		/* Send a RESET to the peer. */
+		lb_peer_reset(lbp);
+		return;
+
+	case LB_PEER_EV_UNAVAILABLE:
+		lb_peer_discard_all_conns(lbp);
+		lb_peer_state_chg(lbp, LB_PEER_ST_WAIT_RX_RESET);
+		return;
+
 	default:
 		LOG_LB_PEER(lbp, LOGL_ERROR, "Unhandled event: %s\n", osmo_fsm_event_name(&lb_peer_fsm, event));
 		return;
 	}
 }
 
+static void lb_peer_st_ready_onleave(struct osmo_fsm_inst *fi, uint32_t next_state)
+{
+	if (next_state != LB_PEER_ST_READY)
+		osmo_stat_item_dec(osmo_stat_item_group_get_item(g_smlc->statg, SMLC_STAT_LB_PEERS_ACTIVE), 1);
+}
+
 static int lb_peer_fsm_timer_cb(struct osmo_fsm_inst *fi)
 {
 	struct lb_peer *lbp = fi->priv;
@@ -336,10 +398,17 @@ static int lb_peer_fsm_timer_cb(struct osmo_fsm_inst *fi)
 	return 0;
 }
 
-void lb_peer_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
+/* struct lb_peer destructor: */
+static void lb_peer_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
 {
 	struct lb_peer *lbp = fi->priv;
+
 	lb_peer_discard_all_conns(lbp);
+
+	if (lbp->fi->state == LB_PEER_ST_READY)
+		osmo_stat_item_dec(osmo_stat_item_group_get_item(g_smlc->statg, SMLC_STAT_LB_PEERS_ACTIVE), 1);
+	osmo_stat_item_dec(osmo_stat_item_group_get_item(g_smlc->statg, SMLC_STAT_LB_PEERS_TOTAL), 1);
+
 	llist_del(&lbp->entry);
 }
 
@@ -352,8 +421,8 @@ static const struct value_string lb_peer_fsm_event_names[] = {
 	OSMO_VALUE_STRING(LB_PEER_EV_MSG_DOWN_CO),
 	OSMO_VALUE_STRING(LB_PEER_EV_RX_RESET),
 	OSMO_VALUE_STRING(LB_PEER_EV_RX_RESET_ACK),
-	OSMO_VALUE_STRING(LB_PEER_EV_CONNECTION_SUCCESS),
-	OSMO_VALUE_STRING(LB_PEER_EV_CONNECTION_TIMEOUT),
+	OSMO_VALUE_STRING(LB_PEER_EV_AVAILABLE),
+	OSMO_VALUE_STRING(LB_PEER_EV_UNAVAILABLE),
 	{}
 };
 
@@ -367,7 +436,8 @@ static const struct osmo_fsm_state lb_peer_fsm_states[] = {
 			| S(LB_PEER_EV_RX_RESET)
 			| S(LB_PEER_EV_MSG_UP_CO_INITIAL)
 			| S(LB_PEER_EV_MSG_UP_CO)
-			| S(LB_PEER_EV_CONNECTION_TIMEOUT)
+			| S(LB_PEER_EV_AVAILABLE)
+			| S(LB_PEER_EV_UNAVAILABLE)
 			,
 		.out_state_mask = 0
 			| S(LB_PEER_ST_WAIT_RX_RESET)
@@ -384,7 +454,8 @@ static const struct osmo_fsm_state lb_peer_fsm_states[] = {
 			| S(LB_PEER_EV_RX_RESET_ACK)
 			| S(LB_PEER_EV_MSG_UP_CO_INITIAL)
 			| S(LB_PEER_EV_MSG_UP_CO)
-			| S(LB_PEER_EV_CONNECTION_TIMEOUT)
+			| S(LB_PEER_EV_AVAILABLE)
+			| S(LB_PEER_EV_UNAVAILABLE)
 			,
 		.out_state_mask = 0
 			| S(LB_PEER_ST_WAIT_RX_RESET)
@@ -396,6 +467,8 @@ static const struct osmo_fsm_state lb_peer_fsm_states[] = {
 	[LB_PEER_ST_READY] = {
 		.name = "READY",
 		.action = lb_peer_st_ready,
+		.onenter = lb_peer_st_ready_onenter,
+		.onleave = lb_peer_st_ready_onleave,
 		.in_event_mask = 0
 			| S(LB_PEER_EV_RX_RESET)
 			| S(LB_PEER_EV_MSG_UP_CO_INITIAL)
@@ -403,6 +476,8 @@ static const struct osmo_fsm_state lb_peer_fsm_states[] = {
 			| S(LB_PEER_EV_MSG_DOWN_CO_INITIAL)
 			| S(LB_PEER_EV_MSG_DOWN_CO)
 			| S(LB_PEER_EV_MSG_DOWN_CL)
+			| S(LB_PEER_EV_AVAILABLE)
+			| S(LB_PEER_EV_UNAVAILABLE)
 			,
 		.out_state_mask = 0
 			| S(LB_PEER_ST_WAIT_RX_RESET)
@@ -482,14 +557,3 @@ int lb_peer_up_l2(struct sccp_lb_inst *sli, const struct osmo_sccp_addr *calling
 
 	return osmo_fsm_inst_dispatch(lb_peer->fi, event, &ctx);
 }
-
-void lb_peer_disconnect(struct sccp_lb_inst *sli, uint32_t conn_id)
-{
-	struct lb_conn *lb_conn;
-	llist_for_each_entry(lb_conn, &sli->lb_conns, entry) {
-		if (lb_conn->sccp_conn_id == conn_id) {
-			lb_conn_discard(lb_conn);
-			return;
-		}
-	}
-}
